/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>
Kasim Baybikov, <kmbaybikov@basistech.ru>
Tim Tkachev, <tvtkachev@basistech.ru>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package flipgroup

import (
	"context"
	"strconv"

	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudapi/flipgroup"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/dc"

	log "github.com/sirupsen/logrus"
)

func handleClientIDsUpdate(ctx context.Context, d *schema.ResourceData, c *controller.ControllerCfg, fg *flipgroup.RecordFLIPGroup, warn *dc.Warnings) {
	addedClients := make([]interface{}, 0)
	removedClients := make([]interface{}, 0)

	old_set, new_set := d.GetChange("client_ids")
	oldSlice := old_set.([]interface{})
	newSlice := new_set.([]interface{})

	for _, oldElem := range oldSlice {
		if !containsClient(newSlice, oldElem) {
			removedClients = append(removedClients, oldElem)
		}
	}

	for _, newElem := range newSlice {
		if !containsClient(oldSlice, newElem) {
			addedClients = append(addedClients, newElem)
		}
	}

	log.Debugf("Found client_ids change with %v deletion(s) and %v addition(s) [flipgroupID=%v]", len(removedClients), len(addedClients), fg.ID)

	if len(addedClients) > 0 {
		for _, id := range addedClients {
			req := flipgroup.ComputeAddRequest{
				FLIPGroupID: fg.ID,
				ComputeID:   uint64(id.(int)),
			}

			if _, err := c.CloudAPI().FLIPGroup().ComputeAdd(ctx, req); err != nil {
				warn.Add(err)
			}
		}
	}

	if len(removedClients) > 0 {
		for _, id := range removedClients {
			req := flipgroup.ComputeRemoveRequest{
				FLIPGroupID: fg.ID,
				ComputeID:   uint64(id.(int)),
			}

			if _, err := c.CloudAPI().FLIPGroup().ComputeRemove(ctx, req); err != nil {
				warn.Add(err)
			}
		}
	}
}

func containsClient(set []interface{}, check interface{}) bool {
	for _, elem := range set {
		elemConv := elem.(int)
		checkConv := check.(int)
		if elemConv == checkConv {
			return true
		}
	}

	return false
}

func addComputesAfterCreation(ctx context.Context, warnings *dc.Warnings, c *controller.ControllerCfg, compute_ids []interface{}, flipgroupID uint64) {
	if len(compute_ids) == 0 {
		return
	}

	log.Debugf("Adding %v computes to flipgroup [id=%v]...", len(compute_ids), flipgroupID)

	for _, elem := range compute_ids {
		compute_id := uint64(elem.(int))
		req := flipgroup.ComputeAddRequest{
			FLIPGroupID: flipgroupID,
			ComputeID:   compute_id,
		}

		_, err := c.CloudAPI().FLIPGroup().ComputeAdd(ctx, req)
		if err != nil {
			warnings.Add(err)
		}
	}
}

func utilityFlipgroupCheckPresence(ctx context.Context, d *schema.ResourceData, m interface{}) (*flipgroup.RecordFLIPGroup, error) {
	c := m.(*controller.ControllerCfg)
	req := flipgroup.GetRequest{}

	if d.Id() == "" {
		req.FLIPGroupID = uint64(d.Get("flipgroup_id").(int))
	} else {
		id, _ := strconv.ParseUint(d.Id(), 10, 64)
		req.FLIPGroupID = id
	}

	fg, err := c.CloudAPI().FLIPGroup().Get(ctx, req)
	if err != nil {
		return nil, err
	}

	return fg, err
}

func utilityFlipgroupListCheckPresence(ctx context.Context, d *schema.ResourceData, m interface{}) (*flipgroup.ListFLIPGroups, error) {
	c := m.(*controller.ControllerCfg)
	req := flipgroup.ListRequest{}

	if name, ok := d.GetOk("name"); ok {
		req.Name = name.(string)
	}

	if vinsId, ok := d.GetOk("vins_id"); ok {
		req.VINSID = uint64(vinsId.(int))
	}

	if vinsName, ok := d.GetOk("vins_name"); ok {
		req.VINSName = vinsName.(string)
	}

	if extNetId, ok := d.GetOk("extnet_id"); ok {
		req.ExtNetID = uint64(extNetId.(int))
	}

	if byIp, ok := d.GetOk("by_ip"); ok {
		req.ByIP = byIp.(string)
	}

	if rgId, ok := d.GetOk("rg_id"); ok {
		req.RGID = uint64(rgId.(int))
	}

	if byId, ok := d.GetOk("by_id"); ok {
		req.ByID = uint64(byId.(int))
	}

	if page, ok := d.GetOk("page"); ok {
		req.Page = uint64(page.(int))
	}

	if size, ok := d.GetOk("size"); ok {
		req.Size = uint64(size.(int))
	}

	fg_list, err := c.CloudAPI().FLIPGroup().List(ctx, req)
	if err != nil {
		return nil, err
	}

	return fg_list, err
}
